@@ -14,7 +14,7 @@ from rasterio.features import (
     bounds, geometry_mask, geometry_window, is_valid_geom, rasterize, sieve,
     shapes)
 
-from .conftest import MockGeoInterface
+from .conftest import MockGeoInterface, gdal_version, requires_gdal_lt_35
 
 
 DEFAULT_SHAPE = (10, 10)
@@ -496,7 +496,7 @@ def test_rasterize_geo_interface(geojson_polygon, basic_image_2x2):
 def test_rasterize_geomcollection_no_hole():
     """
     Make sure that bug reported in
-    https://github.com/mapbox/rasterio/issues/1253
+    https://github.com/rasterio/rasterio/issues/1253
     does not recur.  GeometryCollections are flattened to individual parts,
     and should result in no holes where parts overlap.
     """
@@ -519,7 +519,7 @@ def test_rasterize_geomcollection_no_hole():
 def test_rasterize_multipolygon_no_hole():
     """
     Make sure that bug reported in
-    https://github.com/mapbox/rasterio/issues/1253
+    https://github.com/rasterio/rasterio/issues/1253
     does not recur.  MultiPolygons are flattened to individual parts,
     and should result in no holes where parts overlap.
     """
@@ -565,11 +565,14 @@ def test_rasterize_out_image(basic_geometry, basic_image_2x2):
     assert np.array_equal(basic_image_2x2, out)
 
 
-def test_rasterize_invalid_out_dtype(basic_geometry):
+def test_rasterize_int64_out_dtype(basic_geometry):
     """A non-supported data type for out should raise an exception."""
     out = np.zeros(DEFAULT_SHAPE, dtype=np.int64)
-    with pytest.raises(ValueError):
+    if gdal_version.at_least("3.5"):
         rasterize([basic_geometry], out=out)
+    else:
+        with pytest.raises(ValueError):
+            rasterize([basic_geometry], out=out)
 
 
 def test_rasterize_shapes_out_dtype_mismatch(basic_geometry):
@@ -638,13 +641,19 @@ def test_rasterize_default_value_for_none(basic_geometry, basic_image_2x2):
     )
 
 
-def test_rasterize_invalid_default_value(basic_geometry):
+def test_rasterize_int64_default_value(basic_geometry):
     """A default value that requires an int64 should raise an exception."""
-    with pytest.raises(ValueError):
-        rasterize(
-            [basic_geometry], out_shape=DEFAULT_SHAPE,
-            default_value=1000000000000
-        )
+    if gdal_version.at_least("3.5"):
+            rasterize(
+                [basic_geometry], out_shape=DEFAULT_SHAPE,
+                default_value=1000000000000
+            )
+    else:
+        with pytest.raises(ValueError):
+            rasterize(
+                [basic_geometry], out_shape=DEFAULT_SHAPE,
+                default_value=1000000000000
+            )
 
 
 def test_rasterize_fill_value(basic_geometry, basic_image_2x2):
@@ -661,11 +670,17 @@ def test_rasterize_fill_value(basic_geometry, basic_image_2x2):
 
 def test_rasterize_invalid_fill_value(basic_geometry):
     """A fill value that requires an int64 should raise an exception."""
-    with pytest.raises(ValueError):
+    if gdal_version.at_least("3.5"):
         rasterize(
             [basic_geometry], out_shape=DEFAULT_SHAPE, fill=1000000000000,
             default_value=2
         )
+    else:
+        with pytest.raises(ValueError):
+            rasterize(
+                [basic_geometry], out_shape=DEFAULT_SHAPE, fill=1000000000000,
+                default_value=2
+            )
 
 
 def test_rasterize_fill_value_dtype_mismatch(basic_geometry):
@@ -714,59 +729,63 @@ def test_rasterize_value(basic_geometry, basic_image_2x2):
     )
 
 
+@requires_gdal_lt_35
 def test_rasterize_invalid_value(basic_geometry):
     """A shape value that requires an int64 should raise an exception."""
-    with pytest.raises(ValueError) as ex:
+    with pytest.raises(ValueError, match="Values out of range for supported dtypes"):
         rasterize(
             [(basic_geometry, 1000000000000)], out_shape=DEFAULT_SHAPE
         )
 
-    assert 'dtype must be one of' in str(ex.value)
 
-
-def test_rasterize_supported_dtype(basic_geometry):
+@pytest.mark.parametrize(
+    "dtype,default_value",
+    [
+        ("int16", -32768),
+        ("int32", -2147483648),
+        pytest.param(
+            "uint32",
+            4294967295,
+            marks=pytest.mark.xfail(
+                gdal_version.at_least("3.5"), reason="GDAL regression? Works with 3.4.3"
+            ),
+        ),
+        ("uint8", 255),
+        ("uint16", 65535),
+        ("float32", 1.434532),
+        ("float64", -98332.133422114),
+    ],
+)
+def test_rasterize_supported_dtype(dtype, default_value, basic_geometry):
     """Supported data types should return valid results."""
-    supported_types = (
-        ('int16', -32768),
-        ('int32', -2147483648),
-        ('uint8', 255),
-        ('uint16', 65535),
-        ('uint32', 4294967295),
-        ('float32', 1.434532),
-        ('float64', -98332.133422114)
+    truth = np.zeros(DEFAULT_SHAPE, dtype=dtype)
+    truth[2:4, 2:4] = default_value
+
+    result = rasterize(
+        [basic_geometry],
+        out_shape=DEFAULT_SHAPE,
+        default_value=default_value,
+        dtype=dtype,
     )
+    assert np.array_equal(result, truth)
+    assert np.dtype(result.dtype) == np.dtype(truth.dtype)
 
-    for dtype, default_value in supported_types:
-        truth = np.zeros(DEFAULT_SHAPE, dtype=dtype)
-        truth[2:4, 2:4] = default_value
-
-        result = rasterize(
-            [basic_geometry],
-            out_shape=DEFAULT_SHAPE,
-            default_value=default_value,
-            dtype=dtype
-        )
+    result = rasterize([(basic_geometry, default_value)], out_shape=DEFAULT_SHAPE)
+    if np.dtype(dtype).kind == "f":
+        assert np.allclose(result, truth)
+    else:
         assert np.array_equal(result, truth)
-        assert np.dtype(result.dtype) == np.dtype(truth.dtype)
-
-        result = rasterize(
-            [(basic_geometry, default_value)],
-            out_shape=DEFAULT_SHAPE
-        )
-        if np.dtype(dtype).kind == 'f':
-            assert np.allclose(result, truth)
-        else:
-            assert np.array_equal(result, truth)
-        # Since dtype is auto-detected, it may not match due to upcasting
+    # Since dtype is auto-detected, it may not match due to upcasting
 
 
 def test_rasterize_unsupported_dtype(basic_geometry):
     """Unsupported types should all raise exceptions."""
     unsupported_types = (
         ('int8', -127),
-        ('int64', 20439845334323),
         ('float16', -9343.232)
     )
+    if not gdal_version.at_least("3.5"):
+        unsupported_types += (('int64', 20439845334323),)
 
     for dtype, default_value in unsupported_types:
         with pytest.raises(ValueError):
@@ -831,6 +850,46 @@ def test_rasterize_geo_interface_2(geojson_polygon):
     assert rasterize([GeoObj()], out_shape=DEFAULT_SHAPE).sum() == 4
 
 
+def test_rasterize__numpy_coordinates__fail():
+    # https://github.com/rasterio/rasterio/issues/2385
+    shapes = [
+        (
+            {
+                "type": "LineString",
+                "coordinates": np.array(
+                    [
+                        [425596.0123443, 4932971.35636043],
+                        [425598.03434254, 4932966.09916503],
+                        [425592.56573176, 4932963.99585319],
+                        [425590.54373353, 4932969.2530486],
+                        [425596.0123443, 4932971.35636043],
+                    ]
+                ),
+            },
+            2,
+        ),
+        (
+            {
+                "type": "LineString",
+                "coordinates": np.array(
+                    [
+                        [425582.9243515, 4932973.24623693],
+                        [425592.85588065, 4932951.94800393],
+                        [425584.24595668, 4932947.93313045],
+                        [425574.31442752, 4932969.23136344],
+                        [425582.9243515, 4932973.24623693],
+                    ]
+                ),
+            },
+            2,
+        ),
+    ]
+    out = rasterio.features.rasterize(shapes=shapes, out_shape=(100, 100))
+    assert out.shape == (100, 100)
+    # will fail and be filled with 0
+    assert (out == 0).all()
+
+
 def test_shapes(basic_image):
     """Test creation of shapes from pixel values."""
     results = list(shapes(basic_image))
